;; -*- mode: scheme; coding: utf-8 -*-
;; Copyright (c) 2022 quanrong
;; SPDX-License-Identifier: GPL-3.0-or-later
#!chezscheme

(library (chez-csv)
  (export csv-read csv-write make-dialect)
  (import (chezscheme))

  (define (eor-object)
    ;; end-of-row object
    'eor)

  (define (eor-object? obj)
    (equal? obj (eor-object)))

  (define-record-type dialect
    ;; Describe a CSV variant.
    (fields delimiter  ; field delimiter char, default #\,
            quotechar  ; field quotation char, default #\"
            line-termination ; line termination style, either 'unix or 'dos
            skip-initial-space #| whether to skip all spaces at the beginning
            of every field, default #t |#)
    (protocol (lambda (new)
                (lambda (d q lt sis)
                  (assert (and (symbol? lt)
                               (or (symbol=? 'unix lt)
                                   (symbol=? 'dos lt))))
                  (new d q lt sis)))))

  (define *default-dialect* (make-dialect #\, #\" 'unix #t))


  (define csv-read
    ;; input-port [dialect-record] -> vector-of-(vector-of-strings)
    ;; Read CSV from a port and return it parsed as a vector of rows, where
    ;; every row is a vector of strings.
    (case-lambda
      ((port) (csv-read port *default-dialect*))
      ((port dialect)
       (let* ((col-num (find-fields-number port dialect))
              (length (find-csv-length port dialect)))
         (let ((csv-vec (make-vector length)))
           (do ((i 0 (1+ i)))
               ((>= i length) csv-vec)
             (vector-set! csv-vec i
                          (get-csv-row port col-num dialect))))))))
  

  (define (find-csv-length p dialect)
    ;; input-port dialect-record -> int
    ;; Return the number of rows in the CSV.
    (let ((pos (port-position p))
          (qchar (dialect-quotechar dialect))
          (line-term (dialect-line-termination dialect)))
      (let loop ((quoted #f) (length 0) (prev-ch #\x0) (ch (get-char p)))
        (cond ((eof-object? ch)
               (set-port-position! p pos)
               length)

	      ((char=? ch qchar)
               (loop (not quoted) length ch (get-char p)))

	      ((and (char=? ch #\newline)
		    quoted)
               (loop quoted length ch (get-char p)))

	      ((and (char=? ch #\newline)
		    (symbol=? line-term 'unix))
               (loop quoted (1+ length) ch (get-char p)))

	      ((and (char=? ch #\newline)
		    (symbol=? line-term 'dos))
               (if (char=? #\return prev-ch)
                   (loop quoted (1+ length) ch (get-char p))
                   (loop quoted length ch (get-char p))))

	      (else
               (loop quoted length ch (get-char p)))))))

  
  (define (find-fields-number p dialect)
    ;; input-port dialect-record -> int
    ;; Return the number of fields per row in the CSV.
    (let ((pos (port-position p)))
      (let loop ((number 0))
        (let ((field (get-field p dialect)))
          (if (or (eof-object? field) (eor-object? field))
              (begin (set-port-position! p pos)
                     number)
              (loop (1+ number)))))))


  (define (get-csv-row p fields-n dialect)
    ;; input-port char int dialect-record ->
    ;; -> vector-of-strings | eof-object | eor-object
    ;; Read a CSV row and parse it as a vector of strings. If the first
    ;; character encountered is the eof-object, return it.
    (let ((delim (dialect-delimiter dialect))
          (row (make-vector fields-n "")))
      (let loop ((remaining-n fields-n))
        (let ((field (get-field p dialect)))
          (cond ((and (or (eof-object? field) (eor-object? field))
                      (= remaining-n fields-n))
                 field)
                ((or (eof-object? field) (eor-object? field))
                 row)
                ((= remaining-n 0)
                 (vector-set! row (- fields-n remaining-n) field)
                 row)
                (else
                 (vector-set! row (- fields-n remaining-n) field)
                 (loop (- remaining-n 1))))))))
    
    
  (define (get-field p dialect)
    ;; input-port dialect-record -> string | eof-object | eor-object
    ;; Read and return a single field.
    ;; Directly handle only special cases such as skipping beginning
    ;; whitespace and empty fields, and delegate reading the actual body
    ;; of the field to get-field-body.
    (let ((delim (dialect-delimiter dialect))
          (qtch (dialect-quotechar dialect))
          (skip-ini-spc (dialect-skip-initial-space dialect))
          (line-term (dialect-line-termination dialect)))
      (let init  ; to handle EOF and skipping initial space
          ((ch (get-char p)))
        (cond

	 ((eof-object? ch)
          ch)

         ((and skip-ini-spc (char=? #\space ch)) ; initial space
          (init (get-char p)))

         ((char=? delim ch)                      ; empty body
          "")

         ((and (char=? #\newline ch) (symbol=? line-term 'unix))
          (eor-object))

         ((and (char=? #\return ch) (symbol=? line-term 'dos))
          (let ((ch (peek-char p)))
            (if (char=? #\newline ch)
                (begin
		  (get-char p)
		  (eor-object))
                (get-field-body #\return p dialect))))

         (else                       
          (get-field-body ch p dialect))))))

  
  (define (get-field-body previous-ch p dialect)
    ;; char input-port dialect-record -> string | eof-object | eor-object
    ;; Read and return a single field body.
    ;; Called only after get-field has handled the special cases.
    (let* ((qtch (dialect-quotechar dialect))
           (delim (dialect-delimiter dialect))
           (line-term (dialect-line-termination dialect))
           (quoted (char=? qtch previous-ch)))
      (let loop ((quoted quoted)
                 (ch (get-char p))
                 (result (if quoted
                             ""         ; ignore the previous char, as it
                                        ; was the quotechar
                             (string previous-ch))))
        (cond
         ((eof-object? ch)
          result)

	 ((and (not quoted) (char=? delim ch))
          result)

	 ((and quoted (char=? qtch ch))
          (loop (not quoted) (get-char p) result))

	 ((and (char=? #\newline ch) (symbol=? 'unix line-term)
               (not quoted))
          (unget-char p ch)
          result)

	 ((and (char=? #\return ch) (symbol=? 'dos line-term)
               (not quoted))
          (let ((ch2 (peek-char p)))
            (if (char=? #\newline ch2)
                (begin (unget-char p ch)
                       result)
                (let ((result (append-ch result ch)))
		  (get-char p)
                  (loop quoted (get-char p) (append-ch result ch2))))))

	 (else
          (loop quoted (get-char p) (append-ch result ch)))))))
  

  (define csv-write
    ;; output-port vector-of-(vector of strings) [dialect-record] ->
    ;; -> {side effect}
    (case-lambda
      ((port parsed) (csv-write port parsed *default-dialect*))
      ((port parsed dialect)
       (let ((delim (dialect-delimiter dialect))
             (line-term (dialect-line-termination dialect)))

         (define (write-row parsed-row)
           (do ((len (vector-length parsed-row))
                (i 0 (1+ i)))
               ((>= i len) (put-line-term port dialect))
             (when (and (< i len) (> i 0))
               (put-char port delim))
             (put-field port (vector-ref parsed-row i) dialect)))

         (vector-for-each (lambda (row)
                            (write-row row))
                          parsed)
         (flush-output-port port)))))

  
  (define (put-field p field dialect)
    ;; output-port string dialect-record -> {side effect}
    ;; Check if a field needs to be quoted and write it (quoted if needed).
    (let* ((delim (dialect-delimiter dialect))
           (qtch (dialect-quotechar dialect))
           (line-term (dialect-line-termination dialect)))

      (define (needs-quote? ip)
        ;; input-port -> bool
        (let ((ch (get-char ip)))
          (if (eof-object? ch)
              #f
              (if (or (char=? delim ch)
                      (char=? qtch ch)
                      (case line-term
                        ((unix) (char=? #\newline ch))
                        ((dos) (and (char=? #\return ch)
                                    (not (eof-object? (peek-char ip)))
                                    (char=? #\newline (peek-char ip))))))
                  #t
                  (needs-quote? ip)))))

      (if (needs-quote? (open-input-string field))
          (put-string-quoted p field qtch)
          (put-string p field))))

  
  (define (append-ch str ch)
    ;; string char -> string
    (string-append str (string ch)))

  (define (put-line-term p dialect)
    ;; output-port dialect-record -> {side effect}
    (case (dialect-line-termination dialect)
      ((unix) (newline p))
      ((dos) (put-string p "\r\n"))))

  (define (put-string-quoted p string qtch)
    ;; output-port string char -> {side effect}
    (put-char p qtch)
    (put-string p string)
    (put-char p qtch))) 
