#!/usr/bin/env scheme-script
;; -*- mode: scheme; coding: utf-8 -*- !#
;; Copyright (c) 2022 quanrong
;; SPDX-License-Identifier: GPL-3.0-or-later
#!chezscheme

(import (srfi :64 testing)
        (chez-csv) (chezscheme))

(define-syntax lists->vectors
  (syntax-rules ()
    ((_ '(a b ...)) 
     (list->vector (list (list->vector 'a)
			 (list->vector 'b)
			 ...)))))

(define trivial-csv
"col1,col2,col3
this,is,a
test,1,2.0
3,Fo ur,FI5VE
six,s   even,eight\n")

(define trivial-csv-initial-space
"col1, col2, col3
 this,   is,        a
test,  1,               2.0
        3, Fo ur,     FI5VE
 six,     s   even,       eight\n")

(define trivial-csv-trailing-comma
  "col1,col2,col3,
this,is,a,
test,1,2.0,
3,Fo ur,FI5VE,
six,s   even,eight,\n")

(define trivial-parsed
  (lists->vectors '(("col1" "col2" "col3")
		    ("this" "is" "a")
		    ("test" "1" "2.0")
		    ("3" "Fo ur" "FI5VE")
		    ("six" "s   even" "eight"))))

(define trivial-csv-missing-cols
  "col1,col2,col3
this,is,,
test,,2.0
,Fo ur,FI5VE
six,s   even,,\n")

(define trivial-parsed-missing-cols
  (lists->vectors '(("col1" "col2" "col3")
		    ("this" "is" "")
		    ("test" "" "2.0")
		    ("" "Fo ur" "FI5VE")
		    ("six" "s   even" ""))))

(define missing-cols-no-comma
"col1,col2,col3
this,is
test,1,2.0
3,Fo ur
six,s   even\n")

(define parsed-missing-cols-no-comma
  (lists->vectors '(("col1" "col2" "col3")
		    ("this" "is" "")
		    ("test" "1" "2.0")
		    ("3" "Fo ur" "")
		    ("six" "s   even" ""))))

(define quoted-simple "\"col1\",\"col2\",\"col3\"\n")
(define quoted-commas "\",col1\",\"col,2\",\"col3,\"\n")
(define quoted-newlines "col1,\"c
ol
2\",\"
col
3
\"\n")

(define quoted-simple-parsed (lists->vectors '(("col1" "col2" "col3"))))
(define quoted-commas-parsed (lists->vectors '((",col1" "col,2" "col3,"))))
(define quoted-newlines-parsed (lists->vectors
				'(("col1" "c\nol\n2" "\ncol\n3\n"))))

(define dos-end-of-line
  "col1,col2,col3\r\nthi\ns,is,a\r\ntest,one,two\r
\"thr\r\nee\",four,fi\rve\r\n")

(define dos-end-of-line-parsed
  (lists->vectors '(("col1" "col2" "col3")
                    ("thi\ns" "is" "a")
                    ("test" "one" "two")
                    ("thr\r\nee" "four" "fi\rve"))))


(test-begin "read-test")

(test-equal "read trivial csv"
  trivial-parsed
  (let ((op (open-output-string)))
    (put-string op trivial-csv)
    (csv-read (open-input-string (get-output-string op)))))

(test-equal "read trivial csv with trailing commas"
  trivial-parsed
  (let ((op (open-output-string)))
    (put-string op trivial-csv-trailing-comma)
    (csv-read (open-input-string (get-output-string op)))))

(test-equal "read trivial csv with initial space"
  trivial-parsed
  (let ((op (open-output-string)))
    (put-string op trivial-csv-initial-space)
    (csv-read (open-input-string (get-output-string op)))))

(test-equal "read trivial csv with missing cols"
  trivial-parsed-missing-cols
  (let ((op (open-output-string)))
    (put-string op trivial-csv-missing-cols)
    (csv-read (open-input-string (get-output-string op)))))

(test-equal "read trivial csv with missing cols and commas"
  parsed-missing-cols-no-comma
  (let ((op (open-output-string)))
    (put-string op missing-cols-no-comma)
    (csv-read (open-input-string (get-output-string op)))))

(test-equal "read simple quoted csv"
  quoted-simple-parsed
  (let ((op (open-output-string)))
    (put-string op quoted-simple)
    (csv-read (open-input-string (get-output-string op)))))

(test-equal "read csv with quoted commas"
  quoted-commas-parsed
  (let ((op (open-output-string)))
    (put-string op quoted-commas)
    (csv-read (open-input-string (get-output-string op)))))

(test-equal "read csv with quoted newlines"
  quoted-newlines-parsed
  (let ((op (open-output-string)))
    (put-string op quoted-newlines)
    (csv-read (open-input-string (get-output-string op)))))

(test-equal "read csv with DOS-style end of line"
  dos-end-of-line-parsed
  (let ((op (open-output-string)) (dialect (make-dialect #\, #\" 'dos
                                                         #t)))
    (put-string op dos-end-of-line)
    (csv-read (open-input-string (get-output-string op)) dialect)))


(test-end "read-test")



(test-begin "write-test")


(test-equal "write trivial csv"
  trivial-csv
  (let-values (((op g) (open-string-output-port)))
    (csv-write op trivial-parsed)
    (g)))

(test-equal "write csv with commas"
  quoted-commas
  (let-values (((op g) (open-string-output-port)))
    (csv-write op quoted-commas-parsed)
    (g)))

(test-equal "write csv with newlines"
  quoted-newlines
  (let-values (((op g) (open-string-output-port)))
    (csv-write op quoted-newlines-parsed)
    (g)))

(test-equal "write csv with DOS-style line termination"
  dos-end-of-line
  (let-values (((op g) (open-string-output-port)))
    (csv-write op dos-end-of-line-parsed (make-dialect #\, #\" 'dos
						       #t))
    (g)))


(test-end "write-test")


(exit (if (zero? (test-runner-fail-count (test-runner-get))) 0 1))
